'From Squeak 1.31 of Feb 4, 1998 on 21 March 1998 at 11:17:18 pm'!
"Change Set:		Interpreter-WeakClasses
Date:			21 March 1998
Author:			Andreas Raab

This change set contains necessary modifications for the Behavior-WeakClasses change set in the Interpreter. Since a new class format has been used for weak classes, a number of methods have been updated to return the correct values for the new class type.
In addition, this change set defines ObjectMemory>>isWeak: which is a pre-requisite for finalization support in ObjectMemory (see the ObjectMemory-Finalization change set)."
!


!ObjectMemory methodsFor: 'interpreter access' stamp: 'ar 3/21/98 02:38'!
instantiateClass: classPointer indexableSize: size
	| hash header1 header2 cClass byteSize format inc binc header3 hdrSize fillWord newObj sizeHiBits |
"
	NOTE: This method supports the backward-compatible split instSize field of the
	class format word.  The sizeHiBits will go away and other shifts change by 2
	when the split fields get merged in an (incompatible) image change.
"
	self inline: false.
	checkAssertions ifTrue: [
		size < 0 ifTrue: [ self error: 'cannot have a negative indexable field count' ]].

	hash _ self newObjectHash.
	header1 _ self formatOfClass: classPointer. "Low 2 bits are 0"
	sizeHiBits _ (header1 bitAnd: 16r60000) >> 9.
	header1 _ (header1 bitAnd: 16r1FFFF) bitOr: (hash << 17 bitAnd: 16r1FFE0000).
	header2 _ classPointer.
	header3 _ 0.

	cClass _ header1 bitAnd: 16r1F000. "compact class field from format word"
	byteSize _ (header1 bitAnd: 16rFC) + sizeHiBits. "size in bytes -- low 2 bits are 0"
	format _ (header1 >> 8) bitAnd: 16rF.

	format < 8 ifTrue: [
		"Bitmaps and Arrays"
		inc _ size * 4.
	] ifFalse: [
		"Strings and Methods"
		inc _ (size + 3) bitAnd: 16r1FFFFFFC. "round up"
		binc _ 3 - ((size + 3) bitAnd: 3). "odd bytes"
		"low bits of byte size go in format field"
		header1 _ header1 bitOr: (binc << 8).
	].

	(byteSize + inc) > 255 ifTrue: [
		"requires size header word"
		header3 _ byteSize + inc.
		header1 _ header1 - (byteSize bitAnd: 16rFF).  "Clear qsize field"
	] ifFalse: [
		header1 _ header1 + inc.
	].
	byteSize _ byteSize + inc.

	header3 > 0 ifTrue: [
		"requires full header"
		hdrSize _ 3.
	] ifFalse: [
		cClass = 0
			ifTrue: [ hdrSize _ 2 ]
			ifFalse: [ hdrSize _ 1 ].
	].
	format <= 4  "if pointers, fill with nil oop"
		ifTrue: [ fillWord _ nilObj ]
		ifFalse: [ fillWord _ 0 ].

	newObj _ self allocate: byteSize headerSize: hdrSize h1: header1 h2: header2 h3: header3 fill: fillWord.
	^ newObj! !

!ObjectMemory methodsFor: 'header access' stamp: 'ar 3/21/98 00:30'!
formatOf: oop
"       0      no fields
        1      fixed fields only (possibly containing pointers)
        2      indexable fields only (possibly containing pointers)
        3      both fixed and indexable fields (possibly containing pointers)
        4      both fixed and indexable weak fields (possibly containing pointers).

        5      unused
        6      indexable word fields only (no pointers)
        7      unused
 
    8-11      indexable byte fields only (no pointers) (low 2 bits are low 2 bits of size)
   12-15     compiled methods:
                   # of literal oops specified in method header,
                   followed by indexable bytes (same interpretation of low 2 bits as above)
"

	^ ((self baseHeader: oop) >> 8) bitAnd: 16rF! !

!ObjectMemory methodsFor: 'header access' stamp: 'ar 3/21/98 02:38'!
isWeak: oop
	"Answer true if the argument has only weak fields that can hold oops. See comment in formatOf:"
	^ (self formatOf: oop) = 4! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'ar 3/21/98 02:38'!
lastPointerOf: objectPointer
	"Return the byte offset of the last pointer field of the given object. Works with CompiledMethods, as well as ordinary objects. Can be used even when the type bits are not correct."

	| fmt sz methodHeader |
	self inline: true.
	fmt _ self formatOf: objectPointer.
	fmt <= 4 ifTrue: [
		sz _ self sizeBitsOfSafe: objectPointer.
		^ sz - BaseHeaderSize  "all pointers"
	].
	fmt < 12 ifTrue: [ ^0 ].  "no pointers"

	"CompiledMethod: contains both pointers and bytes:"
	methodHeader _ self longAt: objectPointer + BaseHeaderSize.
	^ ((methodHeader >> 10) bitAnd: 16rFF) * 4 + BaseHeaderSize! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'ar 3/21/98 02:38'!
lastPointerWhileForwarding: oop
	"The given object may have its header word in a forwarding block. Find the offset of the last pointer in the object in spite of this obstacle."

	| header fwdBlock fmt size methodHeader |
	self inline: true.
	header _ self longAt: oop.
	(header bitAnd: MarkBit) ~= 0 ifTrue: [
		"oop is forwarded; get its real header from its forwarding table entry"
		fwdBlock _ header bitAnd: AllButMarkBitAndTypeMask.
		checkAssertions ifTrue: [ self fwdBlockValidate: fwdBlock ].
		header _ self longAt: fwdBlock + 4.
	].

	fmt _ (header >> 8) bitAnd: 16rF.
	fmt <= 4 ifTrue: [
		"do sizeBitsOf: using the header we obtained"
		(header bitAnd: TypeMask) = HeaderTypeSizeAndClass
			ifTrue: [ size _ (self sizeHeader: oop) bitAnd: 16rFFFFFFC ]
			ifFalse: [ size _ header bitAnd: 16rFC ].
		^ size - BaseHeaderSize
	].
	fmt < 12 ifTrue: [ ^ 0 ].  "no pointers"

	methodHeader _ self longAt: oop + BaseHeaderSize.
	^ ((methodHeader >> 10) bitAnd: 16rFF) * 4 + BaseHeaderSize! !


!DynamicInterpreter methodsFor: 'object format' stamp: 'ar 3/21/98 02:32'!
fixedFieldsOf: oop format: fmt length: wordLength
"
	NOTE: This code supports the backward-compatible extension to 8 bits of instSize.
	When we revise the image format, it should become...
	^ (classFormat >> 2 bitAnd: 16rFF) - 1
"
	| class classFormat |
	self inline: true.
	((fmt > 4) or: [fmt = 2]) ifTrue: [^ 0].  "indexable fields only"
	fmt < 2 ifTrue: [^ wordLength].  "fixed fields only (zero or more)"
	
	"fmt = 3 or 4: mixture of fixed and indexable fields, so must look at class format word"
	class _ self fetchClassOf: oop.
	classFormat _ self formatOfClass: class.
	^ (classFormat >> 11 bitAnd: 16rC0) + (classFormat >> 2 bitAnd: 16r3F) - 1
! !

!DynamicInterpreter methodsFor: 'array and stream primitives' stamp: 'ar 3/21/98 02:33'!
subscript: array with: index format: fmt
	"Note: This method assumes that the index is within bounds!!"

	self inline: true.
	fmt <= 4 ifTrue: [  "pointer type objects"
		^ self fetchPointer: index - 1 ofObject: array].
	fmt < 8 ifTrue: [  "long-word type objects"
		^ self positive32BitIntegerFor:
			(self fetchWord: index - 1 ofObject: array)
	] ifFalse: [  "byte-type objects"
		^ self integerObjectOf:
			(self fetchByte: index - 1 ofObject: array)
	].! !

!DynamicInterpreter methodsFor: 'array and stream primitives' stamp: 'ar 3/21/98 02:33'!
subscript: array with: index storing: oopToStore format: fmt
	"Note: This method assumes that the index is within bounds!!"

	| valueToStore |
	self inline: true.
	fmt <= 4 ifTrue: [  "pointer type objects"
		self storePointer: index - 1 ofObject: array withValue: oopToStore.
	] ifFalse: [
		fmt < 8 ifTrue: [  "long-word type objects"
			valueToStore _ self positive32BitValueOf: oopToStore.
			successFlag ifTrue:
				[self storeWord: index - 1 ofObject: array withValue: valueToStore].
		] ifFalse: [  "byte-type objects"
			(self isIntegerObject: oopToStore) ifFalse: [successFlag _ false].
			valueToStore _ self integerValueOf: oopToStore.
			((valueToStore >= 0) and: [valueToStore <= 255]) ifFalse: [successFlag _ false].
			successFlag ifTrue:
				[self storeByte: index - 1 ofObject: array withValue: valueToStore].
		].
	].! !

!DynamicInterpreter methodsFor: 'debug support' stamp: 'ar 3/21/98 02:37'!
okayOop: oop
	"Verify that the given oop is legitimate. Check address, header, and size but not class."

	| sz type fmt |
	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	((0 < oop) & (oop < endOfMemory))
		ifFalse: [ self error: 'oop is not a valid address' ].
	((oop \\ 4) = 0)
		ifFalse: [ self error: 'oop is not a word-aligned address' ].
	sz _ self sizeBitsOf: oop.
	(oop + sz) < endOfMemory
		ifFalse: [ self error: 'oop size would make it extend beyond the end of memory' ].

	"header type checks"
	type _ self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self error: 'oop is a free chunk, not an object' ].
	type = HeaderTypeShort ifTrue: [
		(((self baseHeader: oop) >> 12) bitAnd: 16r1F) = 0
			ifTrue:  [ self error: 'cannot have zero compact class field in a short header' ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= 4) and: [(self headerType: oop - 4) = type])
			ifFalse: [ self error: 'class header word has wrong type' ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= 8) and:
		 [(self headerType: oop - 8) = type and:
		 [(self headerType: oop - 4) = type]])
			ifFalse: [ self error: 'class header word has wrong type' ].
	].

	"format check"
	fmt _ self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self error: 'oop has an unknown format type' ].

	"mark and root bit checks"
	((self longAt: oop) bitAnd: 16r20000000) = 0
		ifFalse: [ self error: 'unused header bit 30 is set; should be zero' ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self error: 'mark bit should not be set except during GC' ].
xxx"
	(((self longAt: oop) bitAnd: RootBit) = 1 and:
	 [oop >= youngStart])
		ifTrue: [ self error: 'root bit is set in a young object' ].
	^true! !


!Interpreter methodsFor: 'object format' stamp: 'ar 3/21/98 02:37'!
fixedFieldsOf: oop format: fmt length: wordLength
"
	NOTE: This code supports the backward-compatible extension to 8 bits of instSize.
	When we revise the image format, it should become...
	^ (classFormat >> 2 bitAnd: 16rFF) - 1
"
	| class classFormat |
	self inline: true.
	((fmt > 4) or: [fmt = 2]) ifTrue: [^ 0].  "indexable fields only"
	fmt < 2 ifTrue: [^ wordLength].  "fixed fields only (zero or more)"
	
	"fmt = 3 or 4: mixture of fixed and indexable fields, so must look at class format word"
	class _ self fetchClassOf: oop.
	classFormat _ self formatOfClass: class.
	^ (classFormat >> 11 bitAnd: 16rC0) + (classFormat >> 2 bitAnd: 16r3F) - 1
! !

!Interpreter methodsFor: 'array and stream primitives' stamp: 'ar 3/21/98 02:37'!
subscript: array with: index format: fmt
	"Note: This method assumes that the index is within bounds!!"

	self inline: true.
	fmt <= 4 ifTrue: [  "pointer type objects"
		^ self fetchPointer: index - 1 ofObject: array].
	fmt < 8 ifTrue: [  "long-word type objects"
		^ self positive32BitIntegerFor:
			(self fetchWord: index - 1 ofObject: array)
	] ifFalse: [  "byte-type objects"
		^ self integerObjectOf:
			(self fetchByte: index - 1 ofObject: array)
	].! !

!Interpreter methodsFor: 'array and stream primitives' stamp: 'ar 3/21/98 02:38'!
subscript: array with: index storing: oopToStore format: fmt
	"Note: This method assumes that the index is within bounds!!"

	| valueToStore |
	self inline: true.
	fmt <= 4 ifTrue: [  "pointer type objects"
		self storePointer: index - 1 ofObject: array withValue: oopToStore.
	] ifFalse: [
		fmt < 8 ifTrue: [  "long-word type objects"
			valueToStore _ self positive32BitValueOf: oopToStore.
			successFlag ifTrue:
				[self storeWord: index - 1 ofObject: array withValue: valueToStore].
		] ifFalse: [  "byte-type objects"
			(self isIntegerObject: oopToStore) ifFalse: [successFlag _ false].
			valueToStore _ self integerValueOf: oopToStore.
			((valueToStore >= 0) and: [valueToStore <= 255]) ifFalse: [successFlag _ false].
			successFlag ifTrue:
				[self storeByte: index - 1 ofObject: array withValue: valueToStore].
		].
	].! !

!Interpreter methodsFor: 'debug support' stamp: 'ar 3/21/98 02:37'!
okayOop: oop
	"Verify that the given oop is legitimate. Check address, header, and size but not class."

	| sz type fmt |
	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	((0 < oop) & (oop < endOfMemory))
		ifFalse: [ self error: 'oop is not a valid address' ].
	((oop \\ 4) = 0)
		ifFalse: [ self error: 'oop is not a word-aligned address' ].
	sz _ self sizeBitsOf: oop.
	(oop + sz) < endOfMemory
		ifFalse: [ self error: 'oop size would make it extend beyond the end of memory' ].

	"header type checks"
	type _ self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self error: 'oop is a free chunk, not an object' ].
	type = HeaderTypeShort ifTrue: [
		(((self baseHeader: oop) >> 12) bitAnd: 16r1F) = 0
			ifTrue:  [ self error: 'cannot have zero compact class field in a short header' ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= 4) and: [(self headerType: oop - 4) = type])
			ifFalse: [ self error: 'class header word has wrong type' ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= 8) and:
		 [(self headerType: oop - 8) = type and:
		 [(self headerType: oop - 4) = type]])
			ifFalse: [ self error: 'class header word has wrong type' ].
	].

	"format check"
	fmt _ self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self error: 'oop has an unknown format type' ].

	"mark and root bit checks"
	((self longAt: oop) bitAnd: 16r20000000) = 0
		ifFalse: [ self error: 'unused header bit 30 is set; should be zero' ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self error: 'mark bit should not be set except during GC' ].
xxx"
	(((self longAt: oop) bitAnd: RootBit) = 1 and:
	 [oop >= youngStart])
		ifTrue: [ self error: 'root bit is set in a young object' ].
	^true! !


!SystemTracer methodsFor: 'private' stamp: 'ar 3/21/98 00:31'!
formatOf: obj
	"Make the code that tells the format of this object.
	It is like the class's instSpec, but with added low bits for byte size"
"       0=      No pointer fields
        1=      Fixed pointer fields only
        2=      Var pointer fields only
        3=      Fixed and var pointer fields
        4=      both fixed and indexable weak fields

        5=      unused
        6=      var long (bit) fields only
        7=      unused
 
        8-11=   var byte fields only
                low 2 bits are low 2 bits of size **
        12-15   methods -- ie #literals in header, followed by var bytes
                same interpretation of low 2 bits"
	| class spec |
	class _ obj class.
	spec _ class instSpec.    "just use what's there"
	spec < 8 ifTrue: [^ spec]
			ifFalse: ["For byte objects, size = wordSize - spec.lowBits"
					^ spec + (3 - (obj size+3 bitAnd: 3))]! !


